"""Utility functions for the vibromaf model"""

import pickle
from pathlib import Path

from sklearn.pipeline import Pipeline, make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.svm import SVR


def make_vibromaf_pipeline() -> Pipeline:
    """Create the vibromaf SVM regressor pipeline"""
    return make_pipeline(StandardScaler(), SVR(kernel="rbf", C=3000, epsilon=0.1))


def save_model(model, model_path: Path) -> None:
    """Save the model into a file"""
    model_path.write_bytes(pickle.dumps(model))


def load_model(model_path: Path):
    """Load the model from a file"""
    return pickle.loads(model_path.read_bytes())



"""Module human vibrotactile perception"""

import math
from dataclasses import dataclass

import numpy as np


@dataclass(frozen=True)
class PerceptualThreshold:
    """Human perceptual threshold"""

    sampling_frequency: int
    x_scaler: float = 62
    y_scaler: float = 1 / 550
    x_offset: float = 1 - 250 * y_scaler
    y_offset: float = 77

    def calculate(self, block_length: int) -> np.array:
        frequency_grid = np.linspace(0, self.sampling_frequency, 2 * block_length)
        frequency_grid = frequency_grid[0:block_length]

        perceptual_threshold = (
            abs(
                self.x_scaler
                / pow((math.log10(self.x_offset)), 2)
                * np.power(
                    (np.log10(self.y_scaler * frequency_grid + self.x_offset)), 2
                )
            )
            - self.y_offset
        )
        limit = np.argmax(perceptual_threshold > 0)
        perceptual_threshold[limit:] = perceptual_threshold[limit]
        return perceptual_threshold



"""Utility functions for MATLAB files"""

from pathlib import Path
from typing import List, Tuple

import numpy as np
from scipy import io

from vibromaf import config


def load_signal_from_mat(mat_file: Path, signal_name: str) -> np.array:
    """Load .mat file and parse signal from it"""
    mat = io.loadmat(str(mat_file))
    try:
        return mat[signal_name]
    except KeyError:
        raise KeyError(f"Available keys: {mat.keys()}")


def load_data_for_metric(
    metric: str, test_indices: List[int]
) -> Tuple[np.array, np.array]:
    """Load and concatenate the training and test data"""
    vcpwq = load_signal_from_mat(
        config.DATA_PATH / f"{metric}_VCPWQ.mat", f"{metric}_VCPWQ"
    )
    pvcslp = load_signal_from_mat(
        config.DATA_PATH / f"{metric}_PVCSLP.mat", f"{metric}_PVCSLP"
    )
    vpcds = load_signal_from_mat(
        config.DATA_PATH / f"{metric}_VPCDS.mat", f"{metric}_VPCDS"
    )

    train_indices = [
        element for element in range(0, vcpwq.shape[1]) if element not in test_indices
    ]

    return np.concatenate(
        [
            vcpwq[:, train_indices].flatten(),
            pvcslp[:, train_indices].flatten(),
            vpcds[:, train_indices].flatten(),
        ]
    ), np.concatenate(
        [
            vcpwq[:, test_indices].flatten(),
            pvcslp[:, test_indices].flatten(),
            vpcds[:, test_indices].flatten(),
        ]
    )


def split_per_codec(data: np.array, number_of_codecs: int = 3) -> np.array:
    """
    Split the data into equal pieces:
    As we concatenate them per codec this is a split per codec
    """
    return np.split(data, number_of_codecs)


def reshape_per_compression_rate(
    data: np.array, number_of_compression_levels: int = 17
) -> np.array:
    """
    Reshape the data into same compression level per row:
    """
    number_of_columns = int(data.size / number_of_compression_levels)
    return data.reshape((number_of_compression_levels, number_of_columns))



"""Transform module"""

import math
from dataclasses import dataclass
from typing import Callable

import numpy as np
from scipy.fftpack import dct

from vibromaf.signal.perception import PerceptualThreshold
from vibromaf.signal.spectrum import mag2db


def compute_block_dft(block: np.array) -> np.array:
    """Compute DFT spectrum using FFT"""
    block_length = np.size(block)
    extended_block = np.zeros(
        2 * block_length,
    )
    extended_block[0:block_length] = block
    spectrum = mag2db(abs(1 / math.sqrt(block_length) * np.fft.fft(extended_block)))[
        0:block_length
    ]
    return spectrum


def compute_block_dct(block: np.array) -> np.array:
    """Compute block transform using DCT"""
    return mag2db(abs(dct(block, norm="ortho")))


def cut_off_strategy(signal: np.array, block_length: int) -> np.array:
    """Prepare signal for block building by cutting off"""
    num_blocks = int(signal.size / block_length)
    return signal[: num_blocks * block_length]


def zero_padding_strategy(signal: np.array, block_length: int) -> np.array:
    """Prepare signal for block building by padding with zeros"""
    num_padding = block_length * math.ceil(signal.size / block_length) - signal.size
    return np.pad(signal, (0, num_padding), "constant")


@dataclass(frozen=True)
class BlockBuilder:
    """Split signal into blocks"""

    block_length: int
    truncation_strategy: Callable[[np.array], np.array] = cut_off_strategy

    def divide(self, signal: np.array) -> np.array:
        if signal.size < self.block_length:
            raise ValueError(
                "Signal is too short! The signal must be at least as long as the block length."
            )
        signal = self.truncation_strategy(signal, self.block_length)
        return np.reshape(signal, (-1, self.block_length))

    def divide_and_normalize(self, signal: np.array) -> np.array:
        blocks = self.divide(signal)
        means = np.apply_along_axis(np.mean, 1, blocks)
        stds = np.apply_along_axis(np.std, 1, blocks)
        return (blocks - means) / stds


@dataclass(frozen=True)
class PerceptualSpectrumBuilder:
    """Calculate perceptual spectrum"""

    block_builder: BlockBuilder = BlockBuilder(512)
    perceptual_threshold: PerceptualThreshold = PerceptualThreshold(2800)
    block_transform_strategy: Callable[[np.array], np.array] = compute_block_dct

    def compute_perceptual_spectrum(self, signal: np.array) -> np.array:
        blocks = self.block_builder.divide(signal)
        spectrum = np.apply_along_axis(self.block_transform_strategy, 1, blocks)
        return spectrum - self.perceptual_threshold.calculate(
            self.block_builder.block_length
        )



"""Common utilities"""

import numpy as np
from sklearn.metrics import mean_squared_error


def print_metirc(text: str, score: float):
    """Print metric score in predefined format"""
    print(f"{text:40s} {score:.3f}")


def print_mse_and_pc(name: str, y_true, y_pred):
    """Print MSE and Pearson Correlation for signal"""
    mse_test = mean_squared_error(y_true, y_pred)
    cor_test = np.corrcoef(y_true, y_pred)[0, 1]

    print_metirc(f"{name} MSE", mse_test)
    print_metirc(f"{name} PC", cor_test)



"""Spectral Perceptual Quality Index"""

from dataclasses import dataclass

import numpy as np

from vibromaf.signal.spectrum import compute_normalized_spectral_difference
from vibromaf.signal.transform import PerceptualSpectrumBuilder


def spqi(
    distorted: np.array, reference: np.array, eta: float = 0.3, threshold: float = -2.0
) -> float:
    """Wrapper function to calculate the SPQI score

    Parameters
    ------
    * `distorted: np.array` Distorted vibrotactile signal.
    * `reference: np.array` Reference vibrotactile signal.
    * `eta: float` Slope of the mapping function between error and score.
    * `threshold: float` Offset of the mapping function.

    Returns
    -------
    * `float` The SPQI score.
    """
    metric = SPQI(eta, threshold)
    return metric.calculate(distorted, reference)


@dataclass(frozen=True)
class SPQI:
    """Spectral Perceptual Quality Index"""

    eta: float
    threshold: float
    perceptual_spectrum_builder = PerceptualSpectrumBuilder()

    def calculate(self, distorted: np.array, reference: np.array) -> float:
        if np.array_equal(distorted, reference):
            return 1

        ref_perceptual_spectrum = (
            self.perceptual_spectrum_builder.compute_perceptual_spectrum(reference)
        )
        dist_perceptual_spectrum = (
            self.perceptual_spectrum_builder.compute_perceptual_spectrum(distorted)
        )

        norm_perceptual_difference = compute_normalized_spectral_difference(
            ref_perceptual_spectrum, dist_perceptual_spectrum
        )

        block_spqi_scores = self.__compute_block_spqi(norm_perceptual_difference)

        return np.mean(block_spqi_scores)

    def __compute_block_spqi(self, normalized_perceptual_difference: np.array) -> float:
        return (
            1 - np.tanh(self.eta * normalized_perceptual_difference - self.threshold)
        ) / 2



"""Vibrotactile Multi-Method Assessment Fusion"""

from pathlib import Path

import numpy as np

from vibromaf.metrics.snr import nsnr
from vibromaf.metrics.spqi import SPQI
from vibromaf.metrics.stsim import STSIM
from vibromaf.util import model


def vibro_maf(distorted: np.array, reference: np.array, model_path: Path) -> float:
    """Wrapper function to calculate the SPQI score"""
    metric = VibroMAF(model_path, SPQI(0.3, -2.0), STSIM(2 / 3))
    return metric.calculate(distorted, reference)


class VibroMAF:
    """Vibrotactile Multi-Method Assessment Fusion"""

    def __init__(self, model_path: Path, spqi: SPQI, st_sim: STSIM) -> None:
        self.__model = model.load_model(model_path)
        self.__spqi = spqi
        self.__st_sim = st_sim

    def calculate(self, distorted: np.array, reference: np.array) -> float:
        signal_to_noise_ratio = nsnr(distorted, reference)
        st_sim = self.__st_sim.calculate(distorted, reference)
        spqi = self.__spqi.calculate(distorted, reference)
        return float(self.__model.predict([[signal_to_noise_ratio, st_sim, spqi]]))



"""Spectral Temporal SIMilarity"""

from dataclasses import dataclass

import numpy as np

from vibromaf.signal.spectrum import compute_spectral_support
from vibromaf.signal.transform import PerceptualSpectrumBuilder


def st_sim(distorted: np.array, reference: np.array, eta: float = 2 / 3) -> float:
    """Wrapper function to calculate the ST-SIM score

    Parameters
    ------
    * `distorted: np.array` Distorted vibrotactile signal.
    * `reference: np.array` Reference vibrotactile signal.
    * `eta: float` Importance of temporal component compared to spectral component. Should be between 0 and 1.

    Returns
    -------
    * `float` The ST-SIM score.
    """
    metric = STSIM(eta)
    return metric.calculate(distorted, reference)


@dataclass(frozen=True)
class STSIM:
    """Spectral Temporal SIMilarity"""

    eta: float
    perceptual_spectrum_builder = PerceptualSpectrumBuilder()

    def calculate(self, distorted: np.array, reference: np.array) -> float:
        if np.array_equal(distorted, reference):
            return 1

        ref_spectral_support = compute_spectral_support(
            self.perceptual_spectrum_builder.compute_perceptual_spectrum(reference)
        )
        dist_spectral_support = compute_spectral_support(
            self.perceptual_spectrum_builder.compute_perceptual_spectrum(distorted)
        )
        spectral_sim = STSIM.compute_sim(ref_spectral_support, dist_spectral_support)

        ref_normalized_blocks = (
            self.perceptual_spectrum_builder.block_builder.divide_and_normalize(
                reference
            )
        )
        dist_normalized_blocks = (
            self.perceptual_spectrum_builder.block_builder.divide_and_normalize(
                distorted
            )
        )
        temporal_sim = STSIM.compute_sim(ref_normalized_blocks, dist_normalized_blocks)

        return pow(temporal_sim, self.eta) * pow(spectral_sim, 1 - self.eta)

    @staticmethod
    def compute_block_sim(
        reference_block: np.array, distorted_block: np.array
    ) -> float:
        return np.sum(reference_block * distorted_block) / np.sum(
            np.power(reference_block, 2)
        )

    @staticmethod
    def compute_sim(reference: np.array, distorted: np.array) -> float:
        block_sim = np.apply_along_axis(
            STSIM.compute_block_sim, 1, reference, distorted
        )
        return np.mean(block_sim)



"""VibroMAF package configuration"""

from pathlib import Path

import setuptools

ROOT_DIR = Path(__file__).parent
README = ROOT_DIR / "README.md"
CHANGELOG = ROOT_DIR / "CHANGELOG.md"
REQUIREMENTS = ROOT_DIR / "requirements.txt"

setuptools.setup(
    name="vibromaf",
    version="0.0.1",
    author="Markus Hofbauer, Andreas Noll",
    author_email="name.surname@tum.de",
    description="Vibrotactile quality metrics and metric fusion",
    long_description=README.read_text() + "\n" + CHANGELOG.read_text(),
    long_description_content_type="text/markdown",
    url="https://github.com/hofbi/vibromaf",
    include_package_data=True,
    install_requires=REQUIREMENTS.read_text().split(),
    license="MIT",
    zip_safe=False,
    keywords="haptics, vibrotactile quality assessment, signal processing",
    project_urls={
        "Bug Tracker": "https://github.com/hofbi/vibromaf/issues",
        "Discussions": "https://github.com/hofbi/vibromaf/discussions",
        "Documentation": "https://hofbi.github.io/vibromaf",
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    package_dir={"": "vibromaf"},
    packages=setuptools.find_packages(where="vibromaf"),
    python_requires=">=3.7",
)



"""Spectrum module"""

import numpy as np
from scipy.stats import norm


def pow2db(power: np.array) -> np.array:
    """
    Convert power to decibels
    https://de.mathworks.com/help/signal/ref/pow2db.html
    """
    return 10.0 * np.log10(power)


def db2pow(decibel: np.array) -> np.array:
    """
    Convert decibel to power
    https://de.mathworks.com/help/signal/ref/db2pow.html
    """
    return np.power(10.0, decibel / 10.0)


def mag2db(power: np.array) -> np.array:
    """
    Convert magnitude to decibels
    https://de.mathworks.com/help/signal/ref/mag2db.html
    """
    return 2 * pow2db(power)


def signal_energy(signal: np.array) -> np.array:
    """Calculate the signal energy"""
    return np.sum(np.power(signal, 2))


def compute_normalized_spectral_difference(
    reference_spectrum: np.array, distorted_spectrum: np.array
) -> np.array:
    """Compute the normalized difference of two spectrums"""
    difference = np.sum(np.abs(db2pow(reference_spectrum) - db2pow(distorted_spectrum)))
    return pow2db(difference / np.sum(np.abs(db2pow(difference))))


def compute_spectral_support(spectrum: np.array, scale: float = 12) -> np.array:
    """Compute the spectral support of perceptual spectrum using a normal distribution cdf"""
    return np.apply_along_axis(norm.cdf, 1, spectrum, scale=scale)



"""Signal to Noise Ratio"""

import numpy as np

from vibromaf.signal.spectrum import pow2db, signal_energy


def snr(distorted: np.array, reference: np.array) -> float:
    """Calculate the signal-to-noise ratio

    Parameters
    ------
    * `distorted: np.array` Distorted signal.
    * `reference: np.array` Reference signal.

    Returns
    -------
    * `float` The SNR.
    """
    return pow2db(signal_energy(reference) / signal_energy(distorted - reference))


def nsnr(
    distorted: np.array, reference: np.array, normalization_db: float = 75
) -> float:
    """Calculate the normalized signal-to-noise ratio restricted to range of 0 to 1"""
    return max(0.0, min(1.0, snr(distorted, reference) / normalization_db))
